/*	radio.c - M.A.R-T>Y. notification daemon written in C
	Copyright (C) 2023-2024 hoff.industries

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <https://www.gnu.org/licenses/>. */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>

#include <poll.h>
#include <unistd.h>
#include <pthread.h>
#include <string.h>

#include <limits.h>

#include <libnotify/notify.h>
#include <libnotify/notification.h>

#ifndef LIBMMP_AUTHOR
	#include "libmmp.h"
#endif

#ifndef LIBMARTY_AUTHOR
	#include "libmarty.h"
#endif

#ifndef MENDPOINTS_H
	#include "mendpoints.h"
#endif

const int TARGET_PROTOCOL[3] = {2, 0, 0};
const int INSTANCE_VERSION[3] = {3, 2, 2};

const char INSTANCE_NAME[] = "m-cradio";
const char INSTANCE_TITLE[] = "Push Notification Daemon";

struct MartyClientEndpoint Endpoints[] = {
	{
		.endpoint = RADIO_EPT_LOW,
		.name = "low",
		.min = 6,
		.max = 7,
		.isAscii = true
	},
	{
		.endpoint = RADIO_EPT_NORMAL,
		.name = "normal",
		.min = 6,
		.max = 7,
		.isAscii = true
	},
	{
		.endpoint = RADIO_EPT_CRITICAL,
		.name = "critical",
		.min = 6,
		.max = 7,
		.isAscii = true
	},
};

int MaxEndpoint = sizeof(Endpoints) / sizeof(Endpoints[0]);

// map an endpoint index to an urgency level
const int NOTIFICATION_MAP[] = {
	NOTIFY_URGENCY_LOW,
	NOTIFY_URGENCY_NORMAL,
	NOTIFY_URGENCY_CRITICAL
};

void push_notify(char *title, char *message, NotifyUrgency urgency) {
	NotifyNotification * pushNotification = \
		notify_notification_new (title, message, "marty");

	GVariant *entryHint = NULL;

	if (pushNotification == NULL) {
		return;
	}

	 entryHint = g_variant_new_string(INSTANCE_NAME);

	if (entryHint != NULL) {
		notify_notification_set_hint(pushNotification, "desktop-entry", entryHint);
	}

	notify_notification_set_urgency(pushNotification, urgency);
	notify_notification_show (pushNotification, NULL);

	// the entryhint is unref'd by this as well so we
	// dont need to do it explicitly after
	g_object_unref(G_OBJECT(pushNotification));
}

void do_notify(char *title, char *message, NotifyUrgency urgency) {

	int formatSize = strlen(title) + BracketFlavourLen + 2;
	char *formatTitle = (char *)malloc(sizeof(char)*(formatSize));

	if (formatTitle == NULL) {
		push_notify("TITLE ALLOC FAILURE", message, NOTIFY_URGENCY_CRITICAL);
		return;
	}

	if (snprintf(formatTitle, formatSize, "%s %s", BracketFlavour, title) < 0) {
		push_notify("TITLE FORMAT FAILURE", message, NOTIFY_URGENCY_CRITICAL);
		return;
	}

	push_notify(formatTitle, message, urgency);
	free(formatTitle);
}

int main(int argc, char **argv) {
	pthread_t networkThread;
	struct MartyClientData data;
	struct MMStatus status;

	char *title;
	char *message;

	int titleIndex = 0;
	int messageIndex = 0;

	int bytesRead = 0;
	struct MMPacket txPacket;
	struct MMPacket *rxPacket = NULL;

	mmp_arg applicationStatus = MMP_APPLICATION_SUCCESS;

	char hostname[HOST_NAME_MAX];

	int endpointConversion = -1;
	struct MartyClientEndpoint *endpoint = NULL;

	struct MartyMessageHandle *deliveryHandle = NULL;

	printf("%s[%s]%s Starting hoff.industries %s - %s v%d.%d.%d\n",
		LoggerColours[LOG_INFO], MartyFlavour, ClearColour, INSTANCE_TITLE,
			INSTANCE_NAME, INSTANCE_VERSION[0], INSTANCE_VERSION[1], INSTANCE_VERSION[2]);

	if (marty_client_data_factory(&data) == -1) {
		marty_part_log("Client data init failure!\n", LOG_CRITICAL);
		exit(EXIT_ERR_CDATA);
	}

	mmstatus_factory(&status);
	mmpacket_factory(&txPacket, true);

	data.versionString = alloc_marty_version_string(INSTANCE_VERSION, INSTANCE_NAME);
	if (data.versionString == NULL) {
		marty_part_log("Version string allocation failure!\n", LOG_CRITICAL);
		exit(EXIT_ERR_VSTRING);
	}

	if (gethostname(hostname, HOST_NAME_MAX) == -1) {
		marty_part_log("Hostname get failure!\n", LOG_CRITICAL);
		exit(EXIT_ERR_HOSTNAME);
	}

	data.connType = CONN_UNIX;
	data.initParameters.auth = "";
	data.initParameters.priority = 0;
	data.initParameters.host = hostname;
	data.initParameters.name = "radio";

	notify_init (data.versionString);

	if (spawn_marty_socket(&data) != MMP_ERR_NONE) {
		marty_full_log("Socket spawn failure!\n", LOG_CRITICAL, &data);
		do_notify(data.versionString, "CRITICAL: Socket spawn failure!", NOTIFY_URGENCY_CRITICAL);
		exit(EXIT_ERR_SPAWN);
	}

	if (marty_init(&data) != MMP_ERR_NONE) {
		marty_full_log("MMP initialisation failure!\n", LOG_CRITICAL, &data);
		do_notify(data.versionString, "CRITICAL: MMP initialisation failure!", NOTIFY_URGENCY_CRITICAL);
		exit(EXIT_ERR_INIT);
	};

	marty_full_log("MMP initialisation success!\n", LOG_STARTUP, &data);

	if (pthread_create(&networkThread, NULL, marty_network_loop, (void *)&data) != 0) {

		marty_full_log("Error spawning network thread!\n", LOG_CRITICAL, &data);
		exit(EXIT_ERR_SPAWN);
	}
	marty_full_log("Network loop started.\n", LOG_STARTUP, &data);

	bytesRead = snprintf(NULL, 0, "I have initialised with priority %d, chief.",
		data.myParameters.priority) + 1;

	message = (char *)malloc(sizeof(char)*bytesRead);

	if (message == NULL || snprintf(message, bytesRead, "I have initialised with priority %d, chief.",
			data.myParameters.priority) < 0) {

		exit(EXIT_ERR_FORMAT);
	}
	do_notify("radio", message, NOTIFY_URGENCY_LOW);

	free(message);
	message = NULL;

	while (true) {

		mmpacket_reset(&txPacket, true);
		applicationStatus = MMP_APPLICATION_SUCCESS;

		if (rxPacket != NULL) {
			mmpacket_reset(rxPacket, true);
			free(rxPacket);
		}

		rxPacket = NULL;

		if (poll(data.pfds, data.nfds, -1) == -1) {
			marty_full_log("Error polling inter-thread pipes!\n", LOG_CRITICAL, &data);
			exit(EXIT_ERR_POLL);
		}

		if (data.pfds[ITC_DELIVERY_PIPE].revents != 0) {
			// check APPLICATION events

			bytesRead = read(data.netDeliveryPipe[0], &deliveryHandle,
				sizeof(struct MartyMessageHandle *));

			if (bytesRead == 0 || bytesRead == -1) {
				marty_full_log("Error reading inter-thread pipes!\n", LOG_CRITICAL, &data);
				exit(EXIT_ERR_ITC);
			}

			if (deliveryHandle == NULL) {

				// if ITC pipe sent -2 means reconnect event occurred.

				bytesRead = snprintf(NULL, 0, "I have re-initialised with priority %d, chief.",
					data.myParameters.priority) + 1;

				message = (char *)malloc(sizeof(char)*bytesRead);

				if (message == NULL || snprintf(message, bytesRead, "I have re-initialised with priority %d, chief.",
						data.myParameters.priority) < 0) {

					exit(EXIT_ERR_FORMAT);
				}

				do_notify("radio", message, NOTIFY_URGENCY_LOW);

				free(message);
				message = NULL;
			}

			continue;
		}

		if (data.pfds[ITC_MESSAGE_PIPE].revents == 0) {
			// since MESSAGE events are last on our list to check
			// we can continue if we don't need to handle
			continue;
		}

		bytesRead = read(data.netMessagePipe[0], &rxPacket,
			sizeof(struct MMPacket *));

		if (bytesRead == 0 || bytesRead == -1) {
			marty_full_log("Error reading inter-thread pipes!\n", LOG_CRITICAL, &data);
			exit(EXIT_ERR_ITC);
		}

		if (rxPacket == NULL) {
			marty_full_log("Received NULL pointer from ITC pipe!\n", LOG_WARNING, &data);
			continue;
		}

		endpointConversion = rxPacket->payloads[PAYL_IDX_ENDPOINT][0];

		if (endpointConversion > MaxEndpoint || endpointConversion < 0) {
			marty_full_log("Requested endpoint outside my range!\n", LOG_WARNING, &data);
			applicationStatus = MMP_APPLICATION_ENDPOINT;

			goto JMP_TX_STATUS;
		}

		endpoint = &Endpoints[endpointConversion];

		if (rxPacket->payloadLen < endpoint->min) {
			marty_full_log("Request payload count less than expected!\n", LOG_WARNING, &data);
			applicationStatus = MMP_APPLICATION_TOO_FEW;

			goto JMP_TX_STATUS;
		}

		if (rxPacket->payloadLen > endpoint->max) {
			marty_full_log("Request payload count more than expected!\n", LOG_WARNING, &data);
			applicationStatus = MMP_APPLICATION_TOO_MANY;

			goto JMP_TX_STATUS;
		}

		titleIndex = 3;
		messageIndex = 5;

		if (rxPacket->payloadLen == 7) {

			titleIndex = 5;
			messageIndex = 6;
		}

		rxPacket->payloads[titleIndex][mmpayloadlen(rxPacket->payloads[titleIndex])] = '\0';
		rxPacket->payloads[messageIndex][mmpayloadlen(rxPacket->payloads[messageIndex])] = '\0';

		title = (char *)rxPacket->payloads[titleIndex];
		message = (char *)rxPacket->payloads[messageIndex];

		do_notify(title, message, NOTIFICATION_MAP[endpointConversion]);

		title = NULL;
		message = NULL;

		JMP_TX_STATUS:

		marty_set_application(&txPacket, &status, rxPacket->payloads[0], applicationStatus);
		if (status.error != MMP_ERR_NONE) {
			marty_full_log("Couldn't set autopacket!\n", LOG_ERROR, &data);
			continue;
		}

		mmp_serialize(&txPacket, &status);
		if (status.error != MMP_ERR_NONE) {
			marty_full_log("Couldn't serialize autopacket!\n", LOG_ERROR, &data);
			continue;
		}

		marty_net_write(&txPacket, &status, &data);
	}
}
